// SortTreeCtrl.cpp : implementation file
// Copyright (c) 2016 by David Wincelberg
//
// All window styles are optional.
// Sorting is selected by a parameter to InsertItem().
// It can also be invoked through calling SortChildren().
/////////////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "SortTreeCtrl.h"

#include "TreeCtrlData.h"
#include "TreeCtrlOrder.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CSortTreeCtrl

// Note:	These parameters relate to the Strategy and Flyweight patterns.
CSortTreeCtrl::CSortTreeCtrl (CTreeCtrlOrder* pTreeCtrlOrder /* = nullptr */,
							  bool bHandleDelete /* = true */)
{
	m_pTreeCtrlOrder = (pTreeCtrlOrder == nullptr ? new CTreeCtrlOrder() : pTreeCtrlOrder);
	m_bHandleDelete  = bHandleDelete;
	m_bAllowOrderMod = true;
}	// CSortTreeCtrl


// Note:	These parameters relate to the Strategy and Flyweight patterns.
CSortTreeCtrl::CSortTreeCtrl (CStringOrder* pStringOrder,
							  bool bHandleDeleteInner /* = true */,
							  bool bHandleDeleteOuter /* = true */)
{
	m_pTreeCtrlOrder = new CTreeCtrlOrder (pStringOrder, bHandleDeleteInner);
	m_bHandleDelete  = bHandleDeleteOuter;
	m_bAllowOrderMod = true;
}	// CSortTreeCtrl


CSortTreeCtrl::~CSortTreeCtrl()
{
	if (m_bHandleDelete)
		delete m_pTreeCtrlOrder;
}


BEGIN_MESSAGE_MAP(CSortTreeCtrl, CTreeCtrl)
	//{{AFX_MSG_MAP(CSortTreeCtrl)
	ON_NOTIFY_REFLECT(TVN_GETDISPINFO, OnGetDispInfo)
	ON_NOTIFY_REFLECT(TVN_DELETEITEM, OnDeleteItem)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CSortTreeCtrl message handlers

// Action:	Replaces the sort-order object.
//			Deletes the old one unless m_bHandleDelete is false.
// Note:	This relates to the Strategy and Flyweight patterns.
void CSortTreeCtrl::SetSortOrder (CTreeCtrlOrder *pTreeCtrlOrder,
								  bool bHandleDelete /* = true */)
{
	if (!m_bAllowOrderMod) {
		TRACE (_T("** Changing the sort-order strategy is not allowed ")
			_T("after adding a string.\n"));
		ASSERT (false);
		return;
	}

	if (pTreeCtrlOrder == nullptr)
		return;

	if (m_bHandleDelete)
		delete m_pTreeCtrlOrder;

	m_pTreeCtrlOrder = pTreeCtrlOrder;
	m_bHandleDelete  = bHandleDelete;
}	// SetSortOrder


// Action:	Replaces the string-order object.
//			Deletes the old one unless CTreeCtrlOrder's m_bHandleDelete is false.
// Note:	This relates to the Strategy and Flyweight patterns.
void CSortTreeCtrl::SetStringOrder (CStringOrder* pStringOrder,
									bool bHandleDelete /*  = true */)
{
	if (!m_bAllowOrderMod) {
		TRACE (_T("** Changing the string-order strategy is not allowed ")
			_T("after adding a string.\n"));
		ASSERT (false);
		return;
	}

	if (pStringOrder == nullptr || m_pTreeCtrlOrder == nullptr)
		return;

	m_pTreeCtrlOrder->SetStringOrder (pStringOrder, bHandleDelete);
}	// SetStringOrder


// Action:	Checks if a treectrl-data object was supplied; o.w. allocates one.
bool CSortTreeCtrl::CheckTreeCtrlData (CTreeCtrlData** ppTreeCtrlData, LPCTSTR lpszItem,
									   LPARAM lParam /* = 0 */)
{
	if (*ppTreeCtrlData == nullptr) {
		try {
			*ppTreeCtrlData = new CTreeCtrlData (lpszItem, lParam);
		}
		catch (CMemoryException* e) {
			e->Delete();
			return false;
		}
	}

	return true;
}	// CheckTreeCtrlData


// Action:	Inserts an item into a tree control and sorts if requested.
HTREEITEM CSortTreeCtrl::InsertItem (LPCTSTR lpszItem, HTREEITEM hParent /* = TVI_ROOT */,
									 HTREEITEM hInsertAfter /* = TVI_LAST */,
									 CTreeCtrlData* pTreeCtrlData /* = nullptr */)
{
	if (lpszItem == nullptr) {
		delete pTreeCtrlData;
		return NULL;
	}

	HTREEITEM hItem;
	bool bSort = false;

	if (!CheckTreeCtrlData (&pTreeCtrlData, lpszItem))
		return NULL;

	if (hInsertAfter == TVI_SORT) {
		hInsertAfter = TVI_LAST;
		bSort = true;
	}

	hItem = CTreeCtrl::InsertItem (LPSTR_TEXTCALLBACK, hParent, hInsertAfter);

	if (hItem != NULL) {
		CTreeCtrl::SetItemData (hItem, reinterpret_cast<DWORD> (pTreeCtrlData));
		if (bSort)
			SortChildren (hParent);
		m_bAllowOrderMod = false;
	}
	else
		delete pTreeCtrlData;

	return hItem;
}	// InsertItem


// Action:	Inserts an item into a tree control and sorts if requested.
HTREEITEM CSortTreeCtrl::InsertItem (LPCTSTR lpszItem, int nImage, int nSelectedImage,
									 HTREEITEM hParent /* = TVI_ROOT */,
									 HTREEITEM hInsertAfter /* = TVI_LAST */,
									 CTreeCtrlData* pTreeCtrlData /* = nullptr */)
{
	if (lpszItem == nullptr) {
		delete pTreeCtrlData;
		return NULL;
	}

	HTREEITEM hItem;
	bool bSort = false;

	if (!CheckTreeCtrlData (&pTreeCtrlData, lpszItem))
		return NULL;

	if (hInsertAfter == TVI_SORT) {
		hInsertAfter = TVI_LAST;
		bSort = true;
	}

	hItem = CTreeCtrl::InsertItem (LPSTR_TEXTCALLBACK, nImage, nSelectedImage,
		hParent, hInsertAfter);

	if (hItem != NULL) {
		CTreeCtrl::SetItemData (hItem, reinterpret_cast<DWORD> (pTreeCtrlData));
		if (bSort)
			SortChildren (hParent);
		m_bAllowOrderMod = false;
	}
	else
		delete pTreeCtrlData;

	return hItem;
}	// InsertItem


// Action:	Inserts an item into a tree control and sorts if requested.
HTREEITEM CSortTreeCtrl::InsertItem (UINT nMask, LPCTSTR lpszItem, int nImage,
									 int nSelectedImage, UINT nState, UINT nStateMask,
									 LPARAM lParam, HTREEITEM hParent /* = TVI_ROOT */,
									 HTREEITEM hInsertAfter /* = TVI_LAST */,
									 CTreeCtrlData* pTreeCtrlData /* = nullptr */)
{
	if (lpszItem == nullptr) {
		delete pTreeCtrlData;
		return NULL;
	}

	HTREEITEM hItem;
	bool bSort = false;

	if (!CheckTreeCtrlData (&pTreeCtrlData, lpszItem, lParam))
		return NULL;

	if (hInsertAfter == TVI_SORT) {
		hInsertAfter = TVI_LAST;
		bSort = true;
	}

	// Valid members: pszText and lParam
	nMask |= (TVIF_TEXT | TVIF_PARAM);

	hItem = CTreeCtrl::InsertItem (nMask, LPSTR_TEXTCALLBACK, nImage,
		nSelectedImage, nState, nStateMask, reinterpret_cast<DWORD> (pTreeCtrlData),
		hParent, hInsertAfter);

	if (hItem != NULL) {
		if (bSort)
			SortChildren (hParent);
		m_bAllowOrderMod = false;
	}
	else
		delete pTreeCtrlData;

	return hItem;
}	// InsertItem


// Action:	Inserts an item into a tree control and sorts if requested.
HTREEITEM CSortTreeCtrl::InsertItem (LPTVINSERTSTRUCT lpInsertStruct,
									 CTreeCtrlData* pTreeCtrlData /* = nullptr */)
{
	if (lpInsertStruct == nullptr || lpInsertStruct->item.pszText == nullptr) {
		delete pTreeCtrlData;
		return NULL;
	}

	HTREEITEM hItem;
	bool bSort = false;

	if (!CheckTreeCtrlData (&pTreeCtrlData, lpInsertStruct->item.pszText,
	lpInsertStruct->item.lParam))
		return NULL;

	// Valid members: pszText and lParam
	lpInsertStruct->item.mask   |= (TVIF_TEXT | TVIF_PARAM);
	lpInsertStruct->item.pszText = LPSTR_TEXTCALLBACK;
	lpInsertStruct->item.lParam  = reinterpret_cast<DWORD> (pTreeCtrlData);

	if (lpInsertStruct->hInsertAfter == TVI_SORT) {
		lpInsertStruct->hInsertAfter = TVI_LAST;
		bSort = true;
	}

	hItem = CTreeCtrl::InsertItem (lpInsertStruct);

	if (hItem != NULL) {
		if (bSort)
			SortChildren (lpInsertStruct->hParent);
		m_bAllowOrderMod = false;
	}
	else
		delete pTreeCtrlData;

	return hItem;
}	// InsertItem


// Action:	Deletes an item from a tree control and
//			may allow the sort and string orders to change.
BOOL CSortTreeCtrl::DeleteItem (HTREEITEM hItem)
{
	BOOL bResult = CTreeCtrl::DeleteItem (hItem);

	if (GetCount() == 0)
		m_bAllowOrderMod = true;

	return bResult;
}	// DeleteItem


// Action:	Fills in the string that is to be displayed.
// Note:	lstrcpyn will null-terminate a string that is too long.
void CSortTreeCtrl::OnGetDispInfo (NMHDR* pNMHDR, LRESULT* pResult)
{
	TV_DISPINFO* pTVDispInfo = reinterpret_cast<TV_DISPINFO*> (pNMHDR);

	if ((pTVDispInfo->item.mask & TVIF_TEXT) != 0
	&&   pTVDispInfo->item.lParam != 0) {
		CTreeCtrlData* pTreeCtrlData = reinterpret_cast<CTreeCtrlData*>
			(pTVDispInfo->item.lParam);

		::lstrcpyn (pTVDispInfo->item.pszText, pTreeCtrlData->GetText(),
			pTVDispInfo->item.cchTextMax);
	}

	*pResult = 0;
}	// OnGetDispInfo


// Action:	Deletes the stored data when an item is deleted.
// Note:	The subclass destructor will be called.
void CSortTreeCtrl::OnDeleteItem (NMHDR* pNMHDR, LRESULT* pResult)
{
	NM_TREEVIEW* pNMTreeView = reinterpret_cast<NM_TREEVIEW*> (pNMHDR);
	delete reinterpret_cast<CTreeCtrlData*> (pNMTreeView->itemOld.lParam);

	*pResult = 0;
}	// OnDeleteItem


// Action:	Clears the tree control and allows the sort and string orders to change.
BOOL CSortTreeCtrl::DeleteAllItems()
{
	m_bAllowOrderMod = true;

	return CTreeCtrl::DeleteAllItems();
}	// DeleteAllItems


// Action:	Compares items based on the string order pointer passed in lParamSort.
int CALLBACK CSortTreeCtrl::CompareFunc (LPARAM lParam1, LPARAM lParam2, LPARAM lParamSort)
{
	CTreeCtrlOrder* pTreeCtrlOrder = reinterpret_cast<CTreeCtrlOrder*> (lParamSort);
	return pTreeCtrlOrder->Compare (lParam1, lParam2);
}	// CompareFunc


// Action:	Sorts based on m_pTreeCtrlOrder.
BOOL CSortTreeCtrl::SortChildren (HTREEITEM hItem)
{
	if (m_pTreeCtrlOrder == nullptr)
		return FALSE;

	TVSORTCB sortCB;

	sortCB.hParent     = hItem;
	sortCB.lpfnCompare = CompareFunc;
	sortCB.lParam      = reinterpret_cast<LPARAM> (m_pTreeCtrlOrder);

	return SortChildrenCB (&sortCB);
}	// SortChildren


// Action:	Stores user data in a CTreeCtrlData object.
BOOL CSortTreeCtrl::SetItemData (HTREEITEM hItem, DWORD dwData)
{
	CTreeCtrlData* pTreeCtrlData = GetItemDataObjPtr (hItem);

	if (pTreeCtrlData != nullptr) {
		pTreeCtrlData->SetLParam (dwData);
		return TRUE;
	}
	else
		return FALSE;
}	// SetItemData


// Action:	Retrieves user data from a CTreeCtrlData object.
DWORD CSortTreeCtrl::GetItemData (HTREEITEM hItem) const
{
	CTreeCtrlData* pTreeCtrlData = GetItemDataObjPtr (hItem);

	if (pTreeCtrlData != nullptr)
		return pTreeCtrlData->GetLParam();
	else
		return 0;
}	// GetItemData


// Action:	Stores user data as a pointer to a CTreeCtrlData object.
// Note:	Copies the data values from the old data object and deletes it.
//			pTreeCtrlData should be from a subclass of CTreeCtrlData.
BOOL CSortTreeCtrl::SetItemDataObjPtr (HTREEITEM hItem, CTreeCtrlData* pTreeCtrlData)
{
	if (pTreeCtrlData == nullptr)
		return FALSE;

	CTreeCtrlData* pOldTreeCtrlData = GetItemDataObjPtr (hItem);

	if (pOldTreeCtrlData != nullptr) {
		pTreeCtrlData = pOldTreeCtrlData;
		delete pOldTreeCtrlData;
		return CTreeCtrl::SetItemData (hItem, reinterpret_cast<DWORD> (pTreeCtrlData));
	}
	else
		return FALSE;
}	// SetItemDataObjPtr


// Action:	Retrieves user data as a pointer to a CTreeCtrlData object.
CTreeCtrlData* CSortTreeCtrl::GetItemDataObjPtr (HTREEITEM hItem) const
{
	if (hItem == NULL)
		return nullptr;

	DWORD dwData = CTreeCtrl::GetItemData (hItem);

	return reinterpret_cast<CTreeCtrlData*> (dwData);
}	// GetItemDataObjPtr
